{ Demo program for ChroMathsome - Linear Algebra Pro.
  Below, you find example implementations of using the functions for eigenvalues
  from Linear Algebra Pro.

  This code was written during the development stages of ChroMathsome - Linear Algebra.
  There's no pre-written script followed so you may find (some of) this code
  (the order of implemented functions) and some of the variables a little strange.

  Note that, since ChroMathsome has its own pre-defined exceptions, you do not need to
  write supersofisticated exception handling routines.
  The only exceptions written here are to make sure the pointers get freed, even
  when an exception has occured (a try ... finally block).

  Since ChroMathsome - Linear Algebra automatically updates the dimensions of
  matrices and vectors (if needed), we gave most of the matrices/vectors in the
  initialization part (InitRealMatrix, ...) a dummy value.
  You are, however, encouraged to enter the correct values (if you know them at forehand)
  since this will make your code much more readable!
  In this demo, we used our privilege to ignore this :-))

  Read the help file for information about the implemented functions
  and their limitations/possibilities!!!

  Author: John Merrit (john.merrit@net4all.be)
  Release date: January 30, 1998
  Reader may do with this code whatever he/she wants.

  Disclaimer:
  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Brought to you by the letter "E".
  "The engine that drives itself - but chooses the uphill climb"              }

unit Eigen;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Menus, StdCtrls, LinAlg, LinAlgP;

type
  TForm1 = class(TForm)
    Memo1: TMemo;
    MainMenu1: TMainMenu;
    Exit1: TMenuItem;
    ClearMemo1: TMenuItem;
    Close1: TMenuItem;
    General1: TMenuItem;
    Roots1: TMenuItem;
    Eigenvectors1: TMenuItem;
    CharEq1: TMenuItem;
    procedure ClearMemo1Click(Sender: TObject);
    procedure Close1Click(Sender: TObject);
    procedure General1Click(Sender: TObject);
    procedure Roots1Click(Sender: TObject);
    procedure CharEq1Click(Sender: TObject);
    procedure Eigenvectors1Click(Sender: TObject);
  private
    { Private declarations }
    procedure Write_Message(s: string);
    procedure Write_RealMatrix(M: RealMatrix);
    procedure Write_ComplexMatrix(M: ComplexMatrix);
    procedure Write_RealVector(V: RealVector);
    procedure Write_ComplexVector(V: ComplexVector);

  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.Write_Message(s: string);
begin
     Memo1.Lines.add(s);
end;

{ Write a real matrix to the screen, creating a nice readable look. }
procedure TForm1.Write_RealMatrix(M: RealMatrix);
var a,b:integer;
    matt,mats,mattt:string;
begin
    memo1.Enabled:= true;
    mattt:= '';
    for a:= 1  to M.Row do
    begin
    mats:= '[ ';
    for b:= 1 to M.Col do
    begin
    if abs(M.Matrix^[a]^[b]) < 1E-17 then
    begin
    M.Matrix^[a]^[b]:= 0;
    matt:= '0';
    end
    else
    matt:= FloatToStrF(M.Matrix^[a]^[b], ffexponent,8,2);
    mattt:= mattt + matt + ' ';
    end;
    mats:= mats + mattt + ']';
    mattt:= '';
    memo1.Lines.add(mats);
    end;
end;

{ Write a complex matrix to the screen, creating a nice readable look. }
procedure TForm1.Write_ComplexMatrix(M: ComplexMatrix);
var a,b:integer;
    matt,mats,mattt:string;
    c,d: extended;
begin
    Memo1.Enabled:= true;
    mattt:= '';
    for a:= 1  to M.Row do
    begin
    mats:= '[ ';
    for b:= 1 to M.Col do
    begin
    c:= M.Matrix^[a]^[b].re;
    if abs(c) < 1E-17 then c:= 0;
    d:= M.Matrix^[a]^[b].im;
    if abs(d) < 1E-17 then d:= 0;
    matt:= '(' + FloatToStr(c) + ';' + FloatToStr(d) + ')';
    mattt:= mattt + matt + ' ';
    end;
    mats:= mats + mattt + ']';
    mattt:= '';
    Memo1.Lines.add(mats);
    end;
end;

{ Write a real vector to the screen, creating a nice readable look. }
procedure TForm1.Write_RealVector(V: RealVector);
var a:integer;
    matt,mats,mattt:string;
begin
    Memo1.Enabled:= true;
    mattt:= '';
    mats:= '[ ';
    for a:= 1 to V.Dimension do
    begin
    if abs(V.Vector^[a]) < 1E-17 then
    begin
    V.Vector^[a]:= 0;
    matt:= IntToStr(round(V.Vector^[a]));
    end
    else
    matt:= FloatToStrF(V.Vector^[a], ffexponent,5,2);
    mattt:= mattt + matt + ' ';
    end;
    mats:= mats + mattt + ']';
    Memo1.Lines.add(mats);
end;

{ Complex counterpart of above. }
procedure TForm1.Write_ComplexVector(V: ComplexVector);
var a:integer;
    matt,mats,mattt:string;
begin
    Memo1.Enabled:= true;
    mattt:= '';
    mats:= '[ ';
    for a:= 1 to V.Dimension do
    begin
    matt:= '(' + FloatToStr(V.Vector^[a].re) + ';' +  FloatToStr(V.Vector^[a].im) + ')';
    mattt:= mattt + matt + ' ';
    end;
    mats:= mats + mattt + ']';
    Memo1.Lines.add(mats);
end;

procedure TForm1.ClearMemo1Click(Sender: TObject);
begin
     if MessageDlg('Delete the contents of this memo?',
     mtInformation, [mbYes, mbNo], 0) = mrYes
     then
     Memo1.Lines.Clear;
end;

procedure TForm1.Close1Click(Sender: TObject);
begin
     Close;
end;

{ Find eigenvalues of a general matrix. }
procedure TForm1.General1Click(Sender: TObject);
var M: RealMatrix;
    V: ComplexVector;
    i: integer;
    c: complex;
    s: string;
begin
     InitRealMatrix(M, 6, 6);
     InitComplexVector(V, 1);

     try
     M.Matrix^[1]^[1]:= 5.3;
     M.Matrix^[1]^[2]:= 2.3;
     M.Matrix^[1]^[3]:= 4.6;
     M.Matrix^[1]^[4]:= 2.7;
     M.Matrix^[1]^[5]:= 1.6;
     M.Matrix^[1]^[6]:= 2.2;
     M.Matrix^[2]^[1]:= 2.4;
     M.Matrix^[2]^[2]:= 7.8;
     M.Matrix^[2]^[3]:= 5.7;
     M.Matrix^[2]^[4]:= 8.4;
     M.Matrix^[2]^[5]:= 3.4;
     M.Matrix^[2]^[6]:= 4.2;
     M.Matrix^[3]^[1]:= 3.4;
     M.Matrix^[3]^[2]:= 5.6;
     M.Matrix^[3]^[3]:= 2.4;
     M.Matrix^[3]^[4]:= 1.7;
     M.Matrix^[3]^[5]:= 7.4;
     M.Matrix^[3]^[6]:= 3.9;
     M.Matrix^[4]^[1]:= 8.3;
     M.Matrix^[4]^[2]:= 7.5;
     M.Matrix^[4]^[3]:= 9.2;
     M.Matrix^[4]^[4]:= 6.1;
     M.Matrix^[4]^[5]:= 5.2;
     M.Matrix^[4]^[6]:= 7.9;
     M.Matrix^[5]^[1]:= 4.3;
     M.Matrix^[5]^[2]:= 5.9;
     M.Matrix^[5]^[3]:= 7.2;
     M.Matrix^[5]^[4]:= 2.6;
     M.Matrix^[5]^[5]:= 4.9;
     M.Matrix^[5]^[6]:= 0.8;
     M.Matrix^[6]^[1]:= 0.9;
     M.Matrix^[6]^[2]:= 2.7;
     M.Matrix^[6]^[3]:= 4.9;
     M.Matrix^[6]^[4]:= 4.8;
     M.Matrix^[6]^[5]:= 6.7;
     M.Matrix^[6]^[6]:= 4.8;

     s:= '';
     Write_Message('Original matrix: ');
     Write_RealMatrix(M);
     RealMatrix_Eigenvalues(V, M);
     Write_Message('Matrix eigenvalues: ');
     for i:= 1 to V.Dimension do
     begin
     s:= 'eigenvalue ' + IntToStr(i) + ': ';
     c:= V.Vector^[i];
     if abs(c.im) < eps then s:= s + FloatToStr(c.re)
     else s:= s + '(' + FloatToStr(c.re) + ';' + FloatToStr(c.im) + ')';
     Write_Message(s);
     end;
     finally
     DestroyRealMatrix(M);
     DestroyComplexVector(V);
     end;
end;

{ Find roots of a polynom. }
procedure TForm1.Roots1Click(Sender: TObject);
var M: RealMatrix;
    V: ComplexVector;
    V1: RealVector;
    i: integer;
begin
     InitRealMatrix(M, 6, 6);
     InitComplexVector(V, 1);
     InitRealVector(V1, 1);

     try
     { Find the eigenvalues for the matrix 5*In.
       The characteristic equation is (5 - )^n.
       In other words, all eigenvalues will be 5 with increasing multiplicity.
       Note that calculating eigenvalues with higher multiplicity is not so obvious as it
       may seem at first sight!
       The higher the multiplicity, the lower the accuracy becomes.
       See also examples lower. }
     for i:= 1 to 10 do
     begin
     Redimension_RealMatrix(M, i, i);
     Redimension_ComplexVector(V, i);
     Redimension_RealVector(V1, i);
     RealIdentityMatrix(M);
     RealMatrix_RealScalarMultiplication(M, 5);
     Write_Message('Original matrix of order ' + IntToStr(i));
     {Write_RealMatrix(M);}
     Write_Message('Eigenvalues: ');
     RealMatrix_Eigenvalues(V, M);
     Write_ComplexVector(V);
     end;

     { Roots of Polynom. See help file for more theoretical background information. }
     { Polynom coefficients -> Frobenius matrix. }
     { Example 1: x^5 - 0.2*x^4 + 7*x^3 + x^2 - 3.5*x + 2 = 0. }
     Redimension_RealVector(V1, 5);
     V1.Vector^[1]:= -0.2;
     V1.Vector^[2]:= 7;
     V1.Vector^[3]:= 1;
     V1.Vector^[4]:= -3.5;
     V1.Vector^[5]:= 2;
     RealFrobeniusMatrix(M, V1);
     Write_Message('Original matrix: ');
     Write_RealMatrix(M);
     RealMatrix_Eigenvalues(V, M);
     Write_Message('Roots of polynom: ');
     Write_ComplexVector(V);

     { Example 2: x^6 - 6*x^5 + 15*x^4 - 20*x^3 + 15*x^2 - 6*x + 1 = 0.
       This example proves the remark about the accuracy, above.
       The roots of this equation are all 1 (thus with a multiplicity of 6).
       The accuracy is only 2 digits!                                        }
     Redimension_RealVector(V1, 6);
     V1.Vector^[1]:= -6;
     V1.Vector^[2]:= 15;
     V1.Vector^[3]:= -20;
     V1.Vector^[4]:= 15;
     V1.Vector^[5]:= -6;
     V1.Vector^[6]:= 1;
     RealFrobeniusMatrix(M, V1);
     Write_Message('Original matrix: ');
     Write_RealMatrix(M);
     RealMatrix_Eigenvalues(V, M);
     Write_Message('Roots of polynom: ');
     Write_ComplexVector(V);
     { Test with multiplicity only 3. }
     Redimension_RealVector(V1, 3);
     V1.Vector^[1]:= -3;
     V1.Vector^[2]:= 3;
     V1.Vector^[3]:= -1;
     RealFrobeniusMatrix(M, V1);
     Write_Message('Original matrix: ');
     Write_RealMatrix(M);
     RealMatrix_Eigenvalues(V, M);
     Write_Message('Roots of polynom: ');
     Write_ComplexVector(V);

     finally
     DestroyRealMatrix(M);
     DestroyComplexVector(V);
     DestroyRealVector(V1);
     end;
end;

{ Find the characteristic equation of a matrix. }
procedure TForm1.CharEq1Click(Sender: TObject);
var M: RealMatrix;
    V1: RealVector;
    i: integer;
begin
     InitRealMatrix(M, 6, 6);
     InitRealVector(V1, 1);

     try
     for i:= 1 to 10 do
     begin
     Redimension_RealMatrix(M, i, i);
     Redimension_RealVector(V1, i);
     Random_RealMatrix(M, 25);
     Write_Message('Original matrix of order ' + IntToStr(i));
     Write_RealMatrix(M);
     RealMatrix_CharacteristicEquation(V1, M);
     Write_Message('Coefficients of characteristic equation: ');
     Write_RealVector(V1);
     end;

     finally
     DestroyRealMatrix(M);
     DestroyRealVector(V1);
     end;
end;

{ Extending ChroMathsome - Linear Algebra.
  As an example, we'll try to find the eigenvectors of a general matrix.
  This example shows two important things:
  a. The ease with which you can extend the functionality of ChroMathsome - Linear Algebra;
  b. The strength of the sysofequations algorithm.
     To find eigenvectors, one needs to solve a homogeneous system of equations (HSE).
     Every HSE has the trivial solution 0.
     Since an eigenvector, by definition, may not be a null vector, we need an algorithm that can
     find non-zero solutions for the HSE.
     Also, since eigenvalues can have higher multiplicities, we need to find (generally, this is not
     always the case cfr. algebraic vs. geometric multiplicities of an eigenvalue) more than one
     eigenvector (or, more theoretically, a basis for an eigenspace associated with an eigenvalue).
  The following code finds all eigenvalues and eigenvectors for some matrices.
  Note: ChroMathsome returns eigenvalues as complex numbers.
  Therefore, we used the sysofequation solver for complex systems.
  If your only interested in real eigenvectors (for real eigenvalues), you can write a simple code
  implementation to use only the real eigenvalues (do this yourself) and use the realsysofequations
  routine.
  In the examples, we ommitted to check for double eigenvalues and did not delete the zero solution
  the sysofequations algorithm gives but since this is only a demonstration ...
  Note: this routine to find eigenvectors may be implemented as a standard function in a future
        version of ChroMathsome - Linear Algebra (code works ok but needs some more polishing ...).  }
procedure TForm1.Eigenvectors1Click(Sender: TObject);
var V, CV: ComplexVector;
    CM, CM2: ComplexMatrix;
    M, M1: RealMatrix;
    i, j: integer;
    eigenval: complex;
    s: string;
begin
     InitRealMatrix(M, 3, 3);
     InitRealMatrix(M1, 3, 3);
     InitComplexMatrix(CM, 1, 1);
     InitComplexMatrix(CM2, 1, 1);
     InitComplexVector(V, M.Row);
     InitComplexVector(CV, M.Row);
     try
     M.Matrix^[1]^[1]:= 1;
     M.Matrix^[1]^[2]:= -3;
     M.Matrix^[1]^[3]:= 3;
     M.Matrix^[2]^[1]:= 3;
     M.Matrix^[2]^[2]:= -5;
     M.Matrix^[2]^[3]:= 3;
     M.Matrix^[3]^[1]:= 6;
     M.Matrix^[3]^[2]:= -6;
     M.Matrix^[3]^[3]:= 4;

     Write_Message('Original matrix: ');
     Write_RealMatrix(M);
     { (M - *In) * X = 0 }
     for i:= 1 to CV.Dimension do
     begin
     CV.Vector^[i].re:= 0;
     CV.Vector^[i].im:= 0;
     end;
     { Find eigenvalues. }
     copy_realmatrix(M1, M);
     RealMatrix_Eigenvalues(V, M1);
     { Find eigenvectors. }
     for i:= 1 to V.Dimension do
     begin
     NilComplexMatrix(CM2);
     RealIdentityMatrix(M1);
     eigenval:= V.Vector^[i];
     RealMatrix_ComplexScalarMultiplication(CM2, M1, eigenval);
     SubtractReal_ComplexMatrix(CM2, M, CM2);
     ComplexSysOfEquations(CM, CM2, CV, j);  { eigenvalues may be complex ... }
     s:= 'eigenvalue ' + IntToStr(i) + ': ';
     if abs(V.Vector^[i].im) = 0 then
     s:= s + FloatToStr(V.Vector^[i].re)
     else s:= s + '(' + FloatToStr(V.Vector^[i].re) + ';' + FloatToStr(V.Vector^[i].im) + ')';
     Write_Message(s);
     { corresponding eigenvector }
     Write_Message('Corresponding eigenvector(s):');
     Write_ComplexMatrix(CM);
     end;

     M.Matrix^[1]^[1]:= 4;
     M.Matrix^[1]^[2]:= 1;
     M.Matrix^[1]^[3]:= -1;
     M.Matrix^[2]^[1]:= 2;
     M.Matrix^[2]^[2]:= 5;
     M.Matrix^[2]^[3]:= -2;
     M.Matrix^[3]^[1]:= 1;
     M.Matrix^[3]^[2]:= 1;
     M.Matrix^[3]^[3]:= 2;

     Write_Message('Original matrix: ');
     Write_RealMatrix(M);
     { (M - *In) * X = 0 }
     for i:= 1 to CV.Dimension do
     begin
     CV.Vector^[i].re:= 0;
     CV.Vector^[i].im:= 0;
     end;
     { Find eigenvalues. }
     copy_realmatrix(M1, M);
     RealMatrix_Eigenvalues(V, M1);
     { Find eigenvectors. }
     for i:= 1 to V.Dimension do
     begin
     NilComplexMatrix(CM2);
     RealIdentityMatrix(M1);
     eigenval:= V.Vector^[i];
     RealMatrix_ComplexScalarMultiplication(CM2, M1, eigenval);
     SubtractReal_ComplexMatrix(CM2, M, CM2);
     ComplexSysOfEquations(CM, CM2, CV, j);  { eigenvalues may be complex ... }
     s:= 'eigenvalue ' + IntToStr(i) + ': ';
     if abs(V.Vector^[i].im) = 0 then
     s:= s + FloatToStr(V.Vector^[i].re)
     else s:= s + '(' + FloatToStr(V.Vector^[i].re) + ';' + FloatToStr(V.Vector^[i].im) + ')';
     Write_Message(s);
     { corresponding eigenvector }
     Write_Message('Corresponding eigenvector(s):');
     Write_ComplexMatrix(CM);
     end;

     M.Matrix^[1]^[1]:= -2;
     M.Matrix^[1]^[2]:= -9;
     M.Matrix^[1]^[3]:= 5;
     M.Matrix^[2]^[1]:= -5;
     M.Matrix^[2]^[2]:= -10;
     M.Matrix^[2]^[3]:= 7;
     M.Matrix^[3]^[1]:= -9;
     M.Matrix^[3]^[2]:= -21;
     M.Matrix^[3]^[3]:= 14;

     Write_Message('Original matrix: ');
     Write_RealMatrix(M);
     { (M - *In) * X = 0 }
     for i:= 1 to CV.Dimension do
     begin
     CV.Vector^[i].re:= 0;
     CV.Vector^[i].im:= 0;
     end;
     { Find eigenvalues. }
     copy_realmatrix(M1, M);
     RealMatrix_Eigenvalues(V, M1);
     { Find eigenvectors. }
     for i:= 1 to V.Dimension do
     begin
     NilComplexMatrix(CM2);
     RealIdentityMatrix(M1);
     eigenval:= V.Vector^[i];
     RealMatrix_ComplexScalarMultiplication(CM2, M1, eigenval);
     SubtractReal_ComplexMatrix(CM2, M, CM2);
     ComplexSysOfEquations(CM, CM2, CV, j);  { eigenvalues may be complex ... }
     s:= 'eigenvalue ' + IntToStr(i) + ': ';
     if abs(V.Vector^[i].im) = 0 then
     s:= s + FloatToStr(V.Vector^[i].re)
     else s:= s + '(' + FloatToStr(V.Vector^[i].re) + ';' + FloatToStr(V.Vector^[i].im) + ')';
     Write_Message(s);
     { corresponding eigenvector }
     Write_Message('Corresponding eigenvector(s):');
     Write_ComplexMatrix(CM);
     end;

     finally
     DestroyRealMatrix(M);
     DestroyRealMatrix(M1);
     DestroyComplexMatrix(CM);
     DestroyComplexMatrix(CM2);
     DestroyComplexVector(V);
     DestroyComplexVector(CV);
     end;
end;

end.
