{ Demo program for ChroMathsome - Linear Algebra.
  Below, you find example implementations of using functions available in
  ChroMathsome - Linear Algebra.

  In this demonstration program we will use matrices to encrypt a text file.

  Here's the pseudo code:
  - form a matrix M where the matrix elements are the ASCII values of the letters of the original text
  - next, we multiply this matrix from the leftside with a random square matrix (decoding matrix)
    => DecM * M
    This results in a matrix ResM which contains the encrypted text.
  - to get the original text back, we multiply the matrix ResM from the leftside
    with the matrix Inverse(DecM)
    => M = Inverse(DecM)*ResM

  Is this a good encryption algorithm?
  Since the decoding matrix (DecM) contains random generated numbers, the chances that
  someone can figure out an algorithm to decode the message is as good as inexisting.
  Note that you must provide the person who needs to retrieve the original text both
  the decoding (or its inverse) matrix and a way to enable him/her to apply the
  matrix calculations (Inverse(DecM)*ResM) as well as an algorithm to read the encrypted message.
  Also note that the dimensions of the matrices increase when you take a larger text file.
  Therefore, I suggest you do not use it to send someone an encrypted copy of your new novel.
  For smaller texts, I think this algorithm is as bullet-proof as any other encryption method
  (if not better!).

  This encryption algorithm is written and invented by the author of this program
  (give credits where credits are due!!!).

  Since ChroMathsome - Linear Algebra automatically updates the dimensions of
  matrices and vectors (if needed), you can give most of the matrices/vectors in the
  initialization part (InitRealMatrix, ...) a dummy value.
  You are, however, encouraged to enter the correct values (if you know them at forehand)
  since this will make your code much more readable!

  Read the help file for information about the implemented functions
  and their limitations/possibilities!

  Author: John Merrit (john.merrit@net4all.be)
  Release date: January 30, 1998
  Reader may do with this code whatever he/she wants.
  (give credits where credits are due ... )

  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Note: before running this demo make sure you know what's going on.
        I've omitted to write this program safe under ALL circumstances.
        It runs fine with the text currently in the memo.
        Since we use matrices to encrypt the text, the more text you wish to encrypt
        the larger the matrices become.
        Also, as the matrices become larger, the accuracy may decrease
        (matrices > 200*200).
        Go ahead and experiment with other texts in the memo
        (try to encrypt and decode this file completely, it works!) ... but do not ask it
        to encrypt a 1000 pages novel
        (unless you got a massive amount of memory in your system, of course).
        On my system I was able to encrypt and decode a 20K file without any problem.
        ( This shows, again, how accurate ChroMathsome - Linear Algebra is!!! )

  Brought to you by the letter "C".              }
unit crypt;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, StdCtrls, LinAlg, LinAlgP;

type

  TForm1 = class(TForm)
    MainMenu1: TMainMenu;
    Memo1: TMemo;
    Close1: TMenuItem;
    Encrypt1: TMenuItem;
    Encrypt2: TMenuItem;
    Decode1: TMenuItem;
    Clear1: TMenuItem;
    procedure Close1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Encrypt2Click(Sender: TObject);
    procedure Decode1Click(Sender: TObject);
    procedure Clear1Click(Sender: TObject);
  private
    { Private declarations }
    M, DecM, ResM: RealMatrix;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.Close1Click(Sender: TObject);
begin
     Close;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
     InitRealMatrix(M, 1, 1);
     InitRealMatrix(DecM, 1, 1);
     InitRealMatrix(ResM, 1, 1);
     Decode1.Enabled:= false;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
     DestroyRealMatrix(M);
     DestroyRealMatrix(DecM);
     DestroyRealMatrix(ResM);
end;

{ Encrypt the text in Memo1. }
procedure TForm1.Encrypt2Click(Sender: TObject);
var i, j, k, max, cnt: integer;
    s: string;
    c: char;
    det: extended;
begin
     max:= 0;
     cnt:= Memo1.Lines.Count;
     if cnt = 0 then
     MessageDlg('There''s no text in the memo to encrypt!', mtError, [mbOK], 0)
     else
     begin
     for i:= 0 to cnt - 1 do
     begin
     s:= Memo1.Lines[i];
     k:= Length(s);
     if k > max then max:= k;
     end;
     Redimension_RealMatrix(M, cnt, max + 1);

     { Fill in matrix elements. }
     for i:= 0 to cnt - 1 do
     begin
     s:= Memo1.Lines[i];
     for j:= 1 to Length(s) + 1 do
     begin
     if j = 1 then M.Matrix^[i+1]^[j]:= Length(s)
     else
     begin
     c:= s[j-1];
     k:= Ord(c);
     M.Matrix^[i+1]^[j]:= k;
     end;
     end;
     end;

     { Encode text. }
     Redimension_RealMatrix(DecM, M.Row, M.Row);  { Make sure the matrix DecM is square! }
     Random_RealMatrix(DecM, 100); { DecM is a random matrix! }
     { Make sure DecM is not singular! }
     Copy_RealMatrix(ResM, DecM);
     det:= RealMatrix_Determinant(ResM);
     if abs(det) < Eps then  { Eps = 1E-18, defined in LinAlg. }
     repeat
     begin
     Random_RealMatrix(DecM, 100);
     Copy_RealMatrix(ResM, DecM);
     det:= RealMatrix_Determinant(ResM);
     end;
     until abs(det) > Eps;
     { Calculate coded matrix. }
     RealMatrix_Multiplication(ResM, DecM, M);

     { Write encrypted text to Memo. }
     Memo1.Lines.Add('Encoded text: ');
     s:= '';
     for j:= 1 to ResM.Row do
     begin
     for k:= 1 to ResM.Col do
     begin
     { Remember that the elements of M are of type extended -> use Round(...) to convert to integer. }
     c:= Chr(Round(ResM.Matrix^[j]^[k]));
     s:= s + c;
     end;
     Memo1.Lines.Add(s);
     s:= '';
     end;
     end;
     Decode1.Enabled:= true;
end;

{ Get the original text back.
  ResM = DecM * M -> M = Inverse(DecM) * ResM }
procedure TForm1.Decode1Click(Sender: TObject);
var i, j, k, max, cnt: integer;
    s: string;
    c: char;
begin
     RealMatrix_Inverse(DecM);
     RealMatrix_Multiplication(M, DecM, ResM);
     Memo1.Lines.Add('Original text: ');
     s:= '';
     for j:= 1 to M.Row do
     begin
     for k:= 2 to M.Col do
     begin
     { If character is newline -> add a space to insert an empty line.
       Remember that the elements of M are of type extended -> use Round(...) to convert to integer. }
     if Round(M.Matrix^[j]^[k]) = 0 then c:= ' ' else
     c:= Chr(Round(M.Matrix^[j]^[k]));
     s:= s + c;
     end;
     Memo1.Lines.Add(s);
     s:= '';
     end;
     NilRealMatrix(M);
     NilRealMatrix(DecM);
     NilRealMatrix(ResM);
     Decode1.Enabled:= false;
end;

procedure TForm1.Clear1Click(Sender: TObject);
begin
     Memo1.Lines.Clear;
end;

end.
